/*
 * Copyright 2010-2011 Craig S. Dickson (http://craigsdickson.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.squeakysand.jsp.tldgenerator;

import com.squeakysand.commons.lang.CompareToHelper;
import com.squeakysand.commons.lang.StringUtils;
import com.squeakysand.commons.lang.ToStringHelper;
import com.squeakysand.schema.javaee.DescriptionType;
import com.squeakysand.schema.javaee.FullyQualifiedClassType;
import com.squeakysand.schema.javaee.FunctionType;
import com.squeakysand.schema.javaee.ObjectFactory;
import com.squeakysand.schema.javaee.TagFileType;
import com.squeakysand.schema.javaee.Taglib;
import com.squeakysand.schema.javaee.TagType;
import com.squeakysand.schema.javaee.TldCanonicalNameType;
import com.squeakysand.schema.javaee.ValidatorType;
import com.squeakysand.schema.javaee.XsdAnyURIType;

import java.net.URI;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class TaglibBuilder {

    private static final Logger LOG = LoggerFactory.getLogger(TaglibBuilder.class);

    private List<TagType> tags;
    private List<FunctionType> functions;
    private List<TagFileType> tagFiles;
    private String shortName;
    private String tlibVersion;
    private URI uri;
    private String validator;
    private String specificationVersion;
    private String description;
    private ObjectFactory objectFactory;

    public TaglibBuilder() {
        objectFactory = new ObjectFactory();
    }

    public void setTags(List<TagType> tags) {
        this.tags = tags;
    }

    public void setFunctions(List<FunctionType> functions) {
        this.functions = functions;
    }

    public void setTagFiles(List<TagFileType> tagFiles) {
        this.tagFiles = tagFiles;
    }

    public void setDescription(String description) {
        this.description = description;
    }

    public void setShortName(String shortName) {
        this.shortName = shortName;
    }

    public void setSpecificationVersion(String specificationVersion) {
        this.specificationVersion = specificationVersion;
    }

    public void setTlibVersion(String tlibVersion) {
        //TODO: validate tlibVersion matches the regular expression from the XSD, check if the pattern is different in different versions
        this.tlibVersion = tlibVersion;
    }

    public void setUri(URI uri) {
        this.uri = uri;
    }

    public void setValidator(String validator) {
        this.validator = validator;
    }

    public Taglib build() {
        Taglib result = objectFactory.createTaglib();
        // set the short name
        if (StringUtils.isNotNullOrEmpty(shortName)) {
            TldCanonicalNameType xmlShortName = objectFactory.createTldCanonicalNameType();
            xmlShortName.setValue(shortName);
            result.setShortName(xmlShortName);
        }
        // set the description
        if (StringUtils.isNotNullOrEmpty(description)) {
            List<DescriptionType> descriptions = result.getDescriptions();
            DescriptionType xmlDescription = objectFactory.createDescriptionType();
            xmlDescription.setValue(description);
            descriptions.add(xmlDescription);
        }
        // set the library version
        if (StringUtils.isNotNullOrEmpty(tlibVersion)) {
            result.setTlibVersion(tlibVersion);
        }
        // set the uri
        if (uri != null) {
            XsdAnyURIType xmlUri = objectFactory.createXsdAnyURIType();
            xmlUri.setValue(uri.toString());
            result.setUri(xmlUri);
        }
        // set the taglib validator
        if (StringUtils.isNotNullOrEmpty(validator)) {
            FullyQualifiedClassType validatorClass = objectFactory.createFullyQualifiedClassType();
            validatorClass.setValue(validator);
            ValidatorType xmlValidator = objectFactory.createValidatorType();
            xmlValidator.setValidatorClass(validatorClass);
            result.setValidator(xmlValidator);
        }
        // set the specification version
        if (StringUtils.isNotNullOrEmpty(specificationVersion)) {
            result.setVersion(specificationVersion);
        }
        // add the tags to the taglib
        if (tags != null) {
            List<TagType> existingTags = result.getTags();
            for (TagType tag : tags) {
                existingTags.add(tag);
            }
            LOG.debug("tags before sorting: {}", ToStringHelper.toString(existingTags));
            Collections.sort(existingTags, new Comparator<TagType>() {

                @Override
                public int compare(TagType t1, TagType t2) {
                    String t1value = t1.getName().getValue();
                    String t2value = t2.getName().getValue();
                    int tempResult = t1value.compareTo(t2value);
                    int result = CompareToHelper.compareTo(t1.getName(), t2.getName(), "value");
                    LOG.debug("tag name compare: {} {} = {} {}", new Object[] {t1value, t2value, tempResult, result});
                    return result;
                }

            });
            LOG.debug("tags after sorting: {}", ToStringHelper.toString(existingTags));
        }
        // add the functions to the taglib
        if (functions != null) {
            List<FunctionType> existingFunctions = result.getFunctions();
            for (FunctionType function : functions) {
                existingFunctions.add(function);
            }
            LOG.debug("functions before sorting: {}", ToStringHelper.toString(existingFunctions));
            Collections.sort(existingFunctions, new Comparator<FunctionType>() {

                @Override
                public int compare(FunctionType f1, FunctionType f2) {
                    return CompareToHelper.compareTo(f1.getName(), f2.getName(), "value");
                }

            });
            LOG.debug("functions after sorting: {}", ToStringHelper.toString(existingFunctions));
        }
        // add the tag files to the taglib
        if (tagFiles != null) {
            List<TagFileType> existingTagFiles = result.getTagFiles();
            for (TagFileType tagFile : tagFiles) {
                existingTagFiles.add(tagFile);
            }
            LOG.debug("tagFiles before sorting: {}", ToStringHelper.toString(existingTagFiles));
            Collections.sort(existingTagFiles, new Comparator<TagFileType>() {

                @Override
                public int compare(TagFileType tf1, TagFileType tf2) {
                    return CompareToHelper.compareTo(tf1.getName(), tf2.getName(), "value");
                }

            });
            LOG.debug("functions after sorting: {}", ToStringHelper.toString(existingTagFiles));
        }
        return result;
    }

}
