/*
 * Copyright 2010-2011 Craig S. Dickson (http://craigsdickson.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.squeakysand.jsp.tldgenerator;

import com.squeakysand.commons.javadoc.JavaDocHelperException;
import com.squeakysand.commons.lang.StringUtils;
import com.squeakysand.commons.javadoc.JavaDocHelper;
import com.squeakysand.jsp.tagext.annotations.JspTagAttribute;
import com.squeakysand.schema.javaee.DescriptionType;
import com.squeakysand.schema.javaee.FullyQualifiedClassType;
import com.squeakysand.schema.javaee.GenericBooleanType;
import com.squeakysand.schema.javaee.JavaIdentifierType;
import com.squeakysand.schema.javaee.ObjectFactory;
import com.squeakysand.schema.javaee.TldAttributeType;

import com.sun.javadoc.AnnotationDesc;
import com.sun.javadoc.ClassDoc;
import com.sun.javadoc.MethodDoc;

import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class TldAttributeTypeListBuilder {

    private static final Logger LOG = LoggerFactory.getLogger(TldAttributeTypeListBuilder.class);

    private ClassDoc classDoc;
    private ObjectFactory objectFactory;

    public TldAttributeTypeListBuilder(ClassDoc classDoc) {
        LOG.debug("created new TldAttributeTypeListBuilder for class: {}", classDoc.qualifiedTypeName());
        this.classDoc = classDoc;
        objectFactory = new ObjectFactory();
    }

    public List<TldAttributeType> build() throws TldGeneratorException {
        LOG.debug("building attributes for class: {}", classDoc.qualifiedTypeName());
        List<TldAttributeType> result = new ArrayList<TldAttributeType>();
        MethodDoc[] methodDocs = classDoc.methods();
        for (MethodDoc methodDoc : methodDocs) {
            if (JavaDocHelper.isJavaBeanSetterMethod(methodDoc)) {
                LOG.debug("{} is a Java bean setter", methodDoc.name());
                AnnotationDesc jspTagAttributeAnnotation = JavaDocHelper.getAnnotation(methodDoc, JspTagAttribute.class.getSimpleName());
                if (jspTagAttributeAnnotation != null) {
                    LOG.debug("{} has JspTagAttribute annotation", methodDoc.name());
                    TldAttributeType attribute = objectFactory.createTldAttributeType();
                    // set the name
                    JavaIdentifierType name = objectFactory.createJavaIdentifierType();
                    try {
                        name.setValue(JavaDocHelper.getJavaBeanPropertyName(methodDoc));
                    } catch (JavaDocHelperException e) {
                        throw new TldGeneratorException(e);
                    }
                    attribute.setName(name);
                    // set the description
                    String attributeDescription = methodDoc.commentText();
                    if (StringUtils.isNotNullOrEmpty(attributeDescription)) {
                        List<DescriptionType> descriptions = attribute.getDescriptions();
                        DescriptionType xmlDescription = objectFactory.createDescriptionType();
                        xmlDescription.setValue(attributeDescription);
                        descriptions.add(xmlDescription);
                    }
                    // check if it is a requred attribute, the default is false
                    Boolean required = (Boolean) JavaDocHelper.getAnnotationElementValue(jspTagAttributeAnnotation, "required");
                    if (required) {
                        GenericBooleanType xmlRequired = objectFactory.createGenericBooleanType();
                        xmlRequired.setValue(required.toString());
                        attribute.setRequired(xmlRequired);
                    }
                    // check if the attribute accepts rt expressions, the default is false
                    Boolean rtexprvalue = (Boolean) JavaDocHelper.getAnnotationElementValue(jspTagAttributeAnnotation, "rtexprvalue");
                    if (rtexprvalue) {
                        GenericBooleanType xmlRtexprvalue = objectFactory.createGenericBooleanType();
                        xmlRtexprvalue.setValue(Boolean.toString(rtexprvalue));
                        attribute.setRtexprvalue(xmlRtexprvalue);
                    }
                    // check to see if the type is something other than a String, also handle fragments
                    String propertyTypeClassName = null;
                    try {
                        propertyTypeClassName = JavaDocHelper.getJavaBeanPropertyTypeName(methodDoc);
                    } catch (JavaDocHelperException e) {
                        throw new TldGeneratorException(e);
                    }
                    if (propertyTypeClassName.equals("javax.servlet.jsp.tagext.JspFragment")) {
                        GenericBooleanType fragment = objectFactory.createGenericBooleanType();
                        fragment.setValue(Boolean.TRUE.toString());
                        attribute.setFragment(fragment);
                    } else if (!propertyTypeClassName.equals("java.lang.String")) {
                        FullyQualifiedClassType type = objectFactory.createFullyQualifiedClassType();
                        type.setValue(propertyTypeClassName);
                        attribute.setType(type);
                    }
                    result.add(attribute);
                } else {
                    LOG.debug("{} does not have JspTagAttribute annotation", methodDoc.name());
                }
            } else {
                LOG.debug("{} skipped, not a Java Bean setter", methodDoc.name());
            }
        }
        return result;
    }

}
