/*
 * Copyright 2010-2011 Craig S. Dickson (http://craigsdickson.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.squeakysand.jsp.tldgenerator;

import com.squeakysand.commons.lang.ToStringHelper;
import com.squeakysand.jsp.tagext.annotations.JspTag;
import com.squeakysand.schema.javaee.FunctionType;
import com.squeakysand.schema.javaee.Taglib;
import com.squeakysand.schema.javaee.TagFileType;
import com.squeakysand.schema.javaee.TagType;

import com.sun.javadoc.AnnotationDesc;
import com.sun.javadoc.ClassDoc;
import com.sun.javadoc.Doclet;
import com.sun.javadoc.DocErrorReporter;
import com.sun.javadoc.LanguageVersion;
import com.sun.javadoc.RootDoc;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class TldGeneratorDoclet extends Doclet {

    private static final Logger LOG = LoggerFactory.getLogger(TldGeneratorDoclet.class);

    private static final Map<String,Integer> VALID_OPTIONS = new HashMap<String,Integer>();

    static {
        // javadoc command options
        VALID_OPTIONS.put("-classpath", 2);
        VALID_OPTIONS.put("-d", 2);
        VALID_OPTIONS.put("-doclet", 2);
        VALID_OPTIONS.put("-docletpath", 2);
        VALID_OPTIONS.put("-encoding", 2);
        VALID_OPTIONS.put("-sourcepath", 2);
        VALID_OPTIONS.put("-source", 2);
        VALID_OPTIONS.put("-verbose", 1);
        // tld generator options
        VALID_OPTIONS.put("-shortName", 2);
        VALID_OPTIONS.put("-description", 2);
        VALID_OPTIONS.put("-tagFilesDir", 2);
        VALID_OPTIONS.put("-tlibVersion", 2);
        VALID_OPTIONS.put("-uri", 2);
        VALID_OPTIONS.put("-specificationVersion", 2);
        VALID_OPTIONS.put("-validator", 2);
        VALID_OPTIONS.put("-includetagfiles", 2);
    }

    public static LanguageVersion languageVersion() {
        return LanguageVersion.JAVA_1_5;
    }

    public static boolean start(RootDoc root) {
        String[][] options = root.options();
        if (LOG.isDebugEnabled()) {
            for (String[] option : options) {
                StringBuilder sb = new StringBuilder();
                for (String val : option) {
                    sb.append(val);
                    sb.append(" ");
                }
                LOG.debug("option: {}", sb.toString());
            }
        }
        boolean result = true;
        try {
            List<TagType> tagTypes = new ArrayList<TagType>();
            List<FunctionType> functionTypes = new ArrayList<FunctionType>();
            ClassDoc[] classDocs = root.classes();
            for (ClassDoc classDoc : classDocs) {
                LOG.debug("processing class: {}", classDoc.name());
                // first parse class as a tag
                if (isAnAnnotatedTagClass(classDoc)) {
                    LOG.debug("{} is a tag class", classDoc.name());
                    File srcDir = new File(getOptionValue("-sourcepath", options));
                    TagTypeBuilder tagTypeBuilder = new TagTypeBuilder(classDoc, srcDir);
                    TagType tagType = tagTypeBuilder.build();
                    tagTypes.add(tagType);
                } else {
                    LOG.debug("{} is not a tag class", classDoc.name());
                }
                // 2nd parse the class looking for function definitions
                FunctionTypeListBuilder functionTypeBuilder = new FunctionTypeListBuilder(classDoc);
                List<FunctionType> functions = functionTypeBuilder.build();
                LOG.debug("found {} functions in class {}", functions.size(), classDoc.qualifiedTypeName());
                functionTypes.addAll(functions);

            }
            // now look for tag files
            List<TagFileType> tagFiles = null;
            Boolean includeTagFiles = Boolean.valueOf(getOptionValue("-includetagfiles", options));
            if (includeTagFiles) {
                File tagFilesDir = getTagFilesDirFromOptions(options);
                TagFileTypeListBuilder tagFileTypeListBuilder = new TagFileTypeListBuilder(tagFilesDir);
                tagFiles = tagFileTypeListBuilder.build();
            }
            TaglibBuilder taglibBuilder = new TaglibBuilder();
            parseOptionsForTaglibProperties(taglibBuilder, options);
            taglibBuilder.setTags(tagTypes);
            taglibBuilder.setFunctions(functionTypes);
            taglibBuilder.setTagFiles(tagFiles);
            Taglib taglibRootElement = taglibBuilder.build();
            File destinationDir = getDestinationDirFromOptions(options);
            TldWriter tldWriter = new TldWriter(destinationDir, taglibRootElement);
            try {
                tldWriter.write();
            } catch (IOException e) {
                LOG.error(e.getMessage(), e);
                result = false;
            } catch (JAXBException e) {
                LOG.error(e.getMessage(), e);
                result = false;
            }
        } catch (TldGeneratorException e) {
            LOG.error(e.getMessage(), e);
        }
        return result;
    }

    public static boolean validOptions(String[][] options, DocErrorReporter reporter) {
        LOG.debug("invoked: options = {}, reporter = {}", ToStringHelper.toString(options), reporter);
        boolean result = true;
        for (String[] option : options) {
            String flag = option[0];
            if (!VALID_OPTIONS.keySet().contains(flag)) {
                result = false;
                reporter.printError(flag + " is not a recognized option");
                break;
            }
        }
        LOG.debug("result: {}", result);
        return result;
    }

    public static int optionLength(String option) {
        LOG.debug("invoked: option = {}", option);
        Integer result = 0;
        result = VALID_OPTIONS.get(option);
        LOG.debug("result: {}", result);
        return result;
    }

    private static boolean isAnAnnotatedTagClass(ClassDoc classDoc) {
        boolean result = false;
        if (classDoc.isClass() && classDoc.isPublic() && !classDoc.isAbstract()) {
            AnnotationDesc[] annotationDescs = classDoc.annotations();
            for (AnnotationDesc annotationDesc : annotationDescs) {
                if (annotationDesc.annotationType().name().equals(JspTag.class.getSimpleName())) {
                    result = true;
                    break;
                }
            }
        }
        return result;
    }

    private static void parseOptionsForTaglibProperties(TaglibBuilder taglibBuilder, String[][] options) throws TldGeneratorException {
        String shortName = getOptionValue("-shortName", options);
        taglibBuilder.setShortName(shortName);
        String tlibVersion = getOptionValue("-tlibVersion", options);
        taglibBuilder.setTlibVersion(tlibVersion);
        String uriPath = getOptionValue("-uri", options);
        try {
            URI uri = new URI(uriPath);
            taglibBuilder.setUri(uri);
        } catch (URISyntaxException e) {
            throw new TldGeneratorException(e);
        }
        String validator = getOptionValue("-validator", options);
        taglibBuilder.setValidator(validator);
        String specificationVersion = getOptionValue("-specificationVersion", options);
        taglibBuilder.setSpecificationVersion(specificationVersion);
        String description = getOptionValue("-description", options);
        taglibBuilder.setDescription(description);
    }

    private static String getOptionValue(String optionName, String[][] options) {
        String result = null;
        for (String[] option : options) {
            String flag = option[0];
            if (flag.equals(optionName)) {
                StringBuilder sb = new StringBuilder();
                for (int i = 1; i < option.length; i++) {
                    sb.append(option[i]);
                    if (i < option.length - 1) {
                        sb.append(" ");
                    }
                }
                result = sb.toString();
            }
        }
        return result;
    }

    private static File getDestinationDirFromOptions(String[][] options) {
        String dirPath = getOptionValue("-d", options);
        return new File(dirPath);
    }

    private static File getTagFilesDirFromOptions(String[][] options) {
        String tagFileDirPath = getOptionValue("-tagFilesDir", options);
        return new File(tagFileDirPath);
    }

}
